"""Defines ValidateColumnStrategy Prototal & Defines a Default Implementation

The Protocol specifies 1 method - validate_column.
The default implementation leverage module level Custom Exceptions

    Typical usage example:

    for c in cols:
            #will raise an error if insufficient
            ValidateColumnStrategyDefault.validate_column_strategy(df, c).validate_column() # noqa: E501
"""
# Standard library imports
import sys

if sys.version_info >= (3, 8):
    # Standard library imports
    from typing import Protocol
else:
    from typing_extensions import Protocol

# Third party imports
import pandas as pd

# chart_me imports
from chart_me.errors import (
    ColumnAllNullError,
    ColumnDoesNotExistsError,
    ColumnTooManyNullsError,
)


class ValidateColumnStrategy(Protocol):
    """Protocol Definition for ValidateColumnStrategy"""

    def validate_column(self) -> bool:
        """Only required method"""
        ...


class ValidateColumnStrategyDefault:
    """Default implentation of ValidateColumnStrategy

    Class is leverage to evaluate whether column selected is viable for charting

    Attributes:
        null_rate: A decimal indicating what percent of null entries in column is ok
    """

    null_rate: float = 0.95  # default check

    def __init__(self, df: pd.DataFrame, col: str, *, override_null_rate: float = 0.95):
        """Init of instance with ability to override class null_rate"""
        self.df = df
        self.col = col
        ValidateColumnStrategyDefault.null_rate = override_null_rate

    def validate_column(self) -> bool:
        """Logic the evaluate column

        Returns:
            True or Raises an Error #TODO think of better strategy

        Raises:
            ColumnDoesNotExistsError
            ColumnAllNullError
            ColumnTooManyNullsError
        """
        # check is exists
        try:
            self.col = self.df[self.col]
        except KeyError:
            raise ColumnDoesNotExistsError(f"{self.col}")

        if self.col.isnull().all():
            raise ColumnAllNullError(f"{self.col}")

        if (
            self.col.isnull().sum() / len(self.col)
            > ValidateColumnStrategyDefault.null_rate
        ):
            raise ColumnTooManyNullsError(
                null_rate=ValidateColumnStrategyDefault.null_rate
            )

        return True



"""contains core chart_me function """
# Standard library imports
from typing import List, Type

# Third party imports
import pandas as pd

# chart_me imports
from chart_me.chart_configs import ChartConfig


def chart_me(
    df: pd.DataFrame, *col_args: str, chart_confs: Type[ChartConfig] = ChartConfig
) -> None:
    """validates, infers data type, assembles and displays charts

    Args:
        df: pandas dataframe
        cols_args: position only collection of column names to render # TODO-(ptrn)
        chart_confs: Manages all settings for Chart creations. Defaults to
            chart_config.ChartConfig -vdefines three classes validate_column_strategy,
            datatype_infer_strategy, assemble_charts_strategy

    Raises:
        ValueError: check number of columns
    """

    cols: List[str] = list(col_args)
    if cols:
        for c in cols:
            # will raise an error if insufficient
            chart_confs.validate_column_strategy(df, c).validate_column()

        # get inferred datatypes
        infer = chart_confs.datatype_infer_strategy(df, cols)
        infer_dtypes = infer.infer_datatypes()

        # get charting strategy
        assembler = chart_confs.assemble_charts_strategy(df, cols, infer_dtypes)
        charts_ = assembler.assemble_charts()
        for c in charts_:
            c.display()

    else:
        raise ValueError("need to send a least one column")



# Standard library imports
from datetime import datetime

# Third party imports
import pandas as pd
import pytest


@pytest.fixture
def conftest_basic_dataframe():
    data = {
        "inty_integers": [1, 2, 3] * 7,
        "floaty_floats": [1.234_523_45, 2.456_234, None] * 7,
        "stringy_strings": ["rabbit", "leopard", None] * 7,
        "datie_dates": [datetime(2020, 1, 1), datetime(2021, 2, 3), None] * 7,
        "all_nulls": [None, None, None] * 7,
        "mostly_nulls": [None, None, None] * 6 + [None, None, "Not Null"],
    }
    df = pd.DataFrame(data)
    return df


@pytest.fixture
def ct_df_check_infer_dtypes():
    data = {
        "inty_integers": [1, 2, 3, 4, 5],
        "inty_integers_w_nulls": [1, 2, 3, None, None],
        "floaty_floats": [1.234_523_45, 2.456_234, 3.5, 4.33, 3.44],
        "stringy_strings": ["rabbit", "leopard", "a", "b", "c"],
        "stringy_strings_w_nulls": ["rabbit", "leopard", "a", None, None],
        "datie_dates": [
            datetime(2020, 1, 1),
            datetime(2021, 2, 3),
            datetime(2021, 2, 13),
            datetime(2021, 4, 13),
            datetime(2021, 3, 13),
        ],
        "datie_dates_null": [
            datetime(2020, 1, 1),
            datetime(2021, 2, 3),
            datetime(2021, 2, 13),
            None,
            None,
        ],
        "boolie_bools": [False, True, True, False, True],
        "boolie_bools_null": [False, None, True, None, None],
    }
    df = pd.DataFrame(data)
    return df


@pytest.fixture
def conftest_check_agg_dataframe():
    data = {
        "unique_vals": [1, 2, 3],
        "unique_with_nulls": [1.234_523_45, 2.456_234, None],
        "duplicates": ["rabbit", "leopard", "rabbit"],
    }
    df = pd.DataFrame(data)
    return df


@pytest.fixture
def conftest_check_agg_dataframe_too_big():
    data = {"unique_vals": list(range(200))}
    df = pd.DataFrame(data)
    return df


@pytest.fixture
def ct_df_override_dtypes():
    data = {
        "stringy_dates": [
            "2021-01-01",
            "2021-02-02",
            "2021-02-03",
            "2021-02-04",
            "2021-02-05",
        ],
        "inty_integers_w_nulls": [1, 2, 3, None, None],
        "floaty_integers": [1.0, 2.0, 3.0, 4.0, 3.0],
        "string_not_dates": [
            "2021-01-01",
            "2021-02-02",
            "2021-02-03",
            "2021-02-04",
            "xyz-yan",
        ],
    }
    df = pd.DataFrame(data)
    return df


@pytest.fixture
def ct_df_calculate_meta_dtypes():
    data = {
        "temporal_t": [
            "2021-01-01",
            "2021-02-02",
            "2021-02-03",
            "2021-02-04",
            "2021-02-05",
        ],
        "integers_b": [0, 1, 1, 0, 1],
        "integers_q": [1.0, 1, 3.0, 4.0, 3.0],
        "integers_k": [1, 2, 3, 4, 5],
        "floaties_q": [1.2, 2.3, 2.2, 4.5, 5.6],
        "nominal_lc": ["A", "B", "B", "A", "B"],
        "nominal_hc": ["A", "A", "B", "C", "D"],
        "nominal_k": ["A", "B", "C", "D", "E"],
    }
    df = pd.DataFrame(data)
    return df



"""A collection of custom errors used in Chart Me

helpful url found here:
https://www.programiz.com/python-programming/user-defined-exception
reminder: default behavior accepts a message
"""


class ColumnDoesNotExistsError(Exception):
    """Implementation to track if column not found in pandas"""

    pass


class ColumnAllNullError(Exception):
    """Implementation of error if column is all nulls"""

    pass


class ColumnTooManyNullsError(Exception):
    """Implementation of error if columns has too many nulls"""

    def __init__(self, null_rate, message="Null Rate below Threshold"):
        """set variables"""
        self.null_rate = null_rate
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        """Override string representations"""
        return f"{self.null_rate} calculated --> {self.message}"


class InsufficientValidColumnsError(Exception):
    """Implementation of error if there's no valid columns to chart"""

    pass



"""Default implementation for bivariate variable charting

This module contains all the logic to go from Metadata to actual charts
See supporting documentation that discusses the rules engine.

    Typical usage example:

    charts = assemble_bivariate_charts(df, [col1, col2], infered_data_types)
"""
# Standard library imports
from typing import List, Union

# Third party imports
import altair as alt
import pandas as pd

# chart_me imports
from chart_me.datatype_infer_strategy import ChartMeDataTypeMetaType, InferedDataTypes
from chart_me.pandas_util import pd_group_me, pd_truncate_date


def assemble_bivariate_charts(
    df: pd.DataFrame, cols: List[str], infered_data_types: InferedDataTypes, **kwargs
) -> List[Union[alt.Chart, alt.HConcatChart]]:
    """Delegated Function to Manage Bivariate Use Cases

    Args:
        df: dataframe
        cols: a list of two columns
        infered_data_types: An instance of InferedDataTypes object

    Returns:
        List of altair charts or compounds charts

    Raises:
        ValueError if called with len of cols != 2
    """
    if len(cols) == 1:
        raise ValueError("This module only supports two valid columns")
    col_name1 = cols[0]
    col_name2 = cols[1]
    merged_meta_types = {
        ChartMeDataTypeMetaType.CATEGORICAL_HIGH_CARDINALITY: ChartMeDataTypeMetaType.KEY,  # noqa: E501
        ChartMeDataTypeMetaType.BOOLEAN: ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY,  # noqa: E501
    }
    col_MT1 = merged_meta_types.get(
        infered_data_types.chart_me_data_types_meta[col_name1],
        infered_data_types.chart_me_data_types_meta[col_name1],
    )
    col_MT2 = merged_meta_types.get(
        infered_data_types.chart_me_data_types_meta[col_name2],
        infered_data_types.chart_me_data_types_meta[col_name2],
    )
    preagg_fl = (
        infered_data_types.preaggregated
    )  # doesn't impact behavior for univariate/bivariate
    return_charts = []

    if set([col_MT1, col_MT2]) == set(
        [ChartMeDataTypeMetaType.QUANTITATIVE, ChartMeDataTypeMetaType.QUANTITATIVE]
    ):
        return_charts.append(build_scatter_plot(df, col_name1, col_name2))

    elif set([col_MT1, col_MT2]) == set(
        [ChartMeDataTypeMetaType.QUANTITATIVE, ChartMeDataTypeMetaType.KEY]
    ):
        col_name_x, col_name_y = (
            [col_name2, col_name1]
            if col_MT1 == ChartMeDataTypeMetaType.KEY
            else [col_name1, col_name2]
        )
        return_charts.append(build_hbar_value(df.head(), col_name_x, col_name_y))

    elif set([col_MT1, col_MT2]) == set(
        [
            ChartMeDataTypeMetaType.QUANTITATIVE,
            ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY,
        ]
    ):
        col_name_n, col_name_q = (
            [col_name1, col_name2]
            if col_MT1 == ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY
            else [col_name2, col_name1]
        )
        if not preagg_fl:
            return_charts.append(build_facet_histogram(df, col_name_n, col_name_q))
        agg_dict = {f"{col_name_q}": ["count", "min", "max", "mean", "median"]}
        df = pd_group_me(df, col_name_n, agg_dict, make_long_form=True)
        # return_charts.insert(0, df)
        # TODO how to store manipulated dataframes to pass back to user
        return_charts.append(
            build_facet_hbars(
                df,
                col_name_facet="measures",
                col_name_y=col_name_n,
                col_name_x=col_name_q,
            )
        )
    elif set([col_MT1, col_MT2]) == set(
        [ChartMeDataTypeMetaType.QUANTITATIVE, ChartMeDataTypeMetaType.TEMPORAL]
    ):
        col_name_t, col_name_q = (
            [col_name1, col_name2]
            if col_MT1 == ChartMeDataTypeMetaType.TEMPORAL
            else [col_name2, col_name1]
        )
        col_name_t_m_y = f"{col_name_t}_m_y"
        df[col_name_t_m_y] = pd_truncate_date(df, col_name_t)
        agg_dict = {col_name_q: ["count", "min", "max", "mean", "median"]}
        df = pd_group_me(
            df, col_name_t_m_y, agg_dict, is_temporal=True, make_long_form=True
        )
        return_charts.append(
            build_hconcat_temp_charts(df, col_name_t_m_y, col_name_q, "measures")
        )
    elif set([col_MT1, col_MT2]) == set(
        [ChartMeDataTypeMetaType.KEY, ChartMeDataTypeMetaType.KEY]
    ):
        df["quantity"] = 1
        df["label"] = df[col_name1].astype(str) + "-" + df[col_name2].astype(str)
        return_charts.append(
            build_hbar_value(df.head(n=20), col_name_x="quantity", col_name_y="label")
        )
    elif set([col_MT1, col_MT2]) == set(
        [
            ChartMeDataTypeMetaType.KEY,
            ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY,
        ]
    ):
        col_name_n, col_name_k = (
            [col_name1, col_name2]
            if col_MT1 == ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY
            else [col_name2, col_name1]
        )
        agg_dict = {col_name_k: ["count", "nunique"]}
        df = pd_group_me(df, col_name_n, agg_dict, make_long_form=True)
        return_charts.append(
            build_facet_hbars(
                df,
                col_name_facet="measures",
                col_name_y=col_name_n,
                col_name_x=col_name_k,
            )
        )
    elif set([col_MT1, col_MT2]) == set(
        [ChartMeDataTypeMetaType.KEY, ChartMeDataTypeMetaType.TEMPORAL]
    ):
        col_name_t, col_name_k = (
            [col_name1, col_name2]
            if col_MT1 == ChartMeDataTypeMetaType.TEMPORAL
            else [col_name2, col_name1]
        )
        col_name_t_m_y = f"{col_name_t}_m_y"
        df[col_name_t_m_y] = pd_truncate_date(df, col_name_t)
        agg_dict = {col_name_k: ["count", "nunique"]}
        df = pd_group_me(
            df, col_name_t_m_y, agg_dict, is_temporal=True, make_long_form=True
        )
        return_charts.append(
            build_hconcat_temp_charts(df, col_name_t_m_y, col_name_k, "measures")
        )
    elif set([col_MT1, col_MT2]) == set(
        [
            ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY,
            ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY,
        ]
    ):
        df["_counts_"] = 1
        df = pd_group_me(
            df,
            cols=[col_name1, col_name2],
            agg_dict={"_counts_": ["sum"]},
            make_long_form=True,
        )
        return_charts.append(build_heatmap(df, col_name1, col_name2, "_counts_"))
    elif set([col_MT1, col_MT2]) == set(
        [
            ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY,
            ChartMeDataTypeMetaType.TEMPORAL,
        ]
    ):
        col_name_t, col_name_n = (
            [col_name1, col_name2]
            if col_MT1 == ChartMeDataTypeMetaType.TEMPORAL
            else [col_name2, col_name1]
        )
        col_name_t_m_y = f"{col_name_t}_m_y"
        df[col_name_t_m_y] = pd_truncate_date(df, col_name_t)
        df["_counts_"] = 1
        df = pd_group_me(
            df,
            cols=[col_name_t_m_y, col_name_n],
            agg_dict={"_counts_": ["sum"]},
            make_long_form=True,
        )
        return_charts.append(
            build_hconcat_temp_lc_charts(df, col_name_t_m_y, col_name_n, "_counts_")
        )
    elif set([col_MT1, col_MT2]) == set(
        [ChartMeDataTypeMetaType.TEMPORAL, ChartMeDataTypeMetaType.TEMPORAL]
    ):
        col_name_t_m_y1, col_name_t_m_y2 = [f"{col_name1}_m_y", f"{col_name2}_m_y"]
        df[col_name_t_m_y1] = pd_truncate_date(df, col_name1)
        df[col_name_t_m_y2] = pd_truncate_date(df, col_name2)
        df["__day_diff__"] = (df[col_name1] - df[col_name2]).dt.days
        # chart_me imports
        from chart_me.charting_assembly_strategy.univariate import build_histogram

        return_charts.append(build_histogram(df, "__day_diff__"))
        df["_counts_"] = 1
        df = pd_group_me(
            df,
            cols=[col_name_t_m_y1, col_name_t_m_y2],
            agg_dict={"_counts_": ["sum"]},
            make_long_form=True,
        )
        return_charts.append(
            build_heatmap(df, col_name_t_m_y1, col_name_t_m_y2, "_counts_")
        )

    else:
        raise NotImplementedError(
            f"unknown handling of metatype-{str(col_MT1)}-{str(col_MT2)}"
        )
    return return_charts


def build_scatter_plot(df: pd.DataFrame, col_name1: str, col_name2: str) -> alt.Chart:
    """An implementation of scatter plot"""
    chart = alt.Chart(df).mark_point().encode(x=f"{col_name1}:Q", y=f"{col_name2}:Q")
    return chart


def build_hbar_value(df: pd.DataFrame, col_name_x: str, col_name_y: str) -> alt.Chart:
    """An implementation of horizontal bar chart"""
    chart = alt.Chart(df).mark_bar().encode(x=f"{col_name_x}:Q", y=f"{col_name_y}:O")
    return chart


def build_facet_histogram(
    df: pd.DataFrame, col_name_facet: str, col_name_hist_q: str
) -> alt.Chart:
    """An implementation of histogram faceted by nominal variable"""
    chart = (
        alt.Chart(df)
        .mark_bar()
        .encode(
            alt.X(f"{col_name_hist_q}:Q", bin=True),
            y="count()",
            facet=alt.Facet(f"{col_name_facet}:O", columns=4),
        )
    )
    return chart


def build_facet_hbars(
    df: pd.DataFrame, col_name_facet: str, col_name_y: str, col_name_x: str
) -> alt.Chart:
    """An implementation of horizontal bar graph faceted by nominal variable"""
    chart = (
        alt.Chart(df)
        .mark_bar()
        .encode(
            x=f"{col_name_x}:Q",
            y=f"{col_name_y}:O",
            facet=alt.Facet(f"{col_name_facet}:O", columns=2),
        )
        .resolve_scale(x="independent")
    )
    return chart


def build_hconcat_temp_charts(
    df: pd.DataFrame, col_name_y_m, col_name_q, col_name_measure: str = "measures"
) -> alt.HConcatChart:
    """An implementation of horizontal bar graph faceted by nominal variable

    WARNING: Function assumes processing through pd_group_me to separate count
    from other aggregations
    """
    df_cnt = df[df[col_name_measure] == "count"]
    df_msr = df[df[col_name_measure] != "count"]

    chart1 = (
        alt.Chart(df_cnt).mark_bar().encode(x=f"{col_name_y_m}:O", y=f"{col_name_q}:Q")
    )
    chart2 = (
        alt.Chart(df_msr)
        .mark_line()
        .encode(x=f"{col_name_y_m}:O", y=f"{col_name_q}:Q", color="measures")
    )
    return alt.hconcat(chart1, chart2)


def build_heatmap(
    df: pd.DataFrame, col_name_x: str, col_name_y: str, col_name_q: str
) -> alt.Chart:
    """An implementation of heatmap"""
    chart = (
        alt.Chart(df)
        .mark_rect()
        .encode(x=f"{col_name_x}:O", y=f"{col_name_y}:O", color=f"{col_name_q}:Q")
    )
    return chart


def build_hconcat_temp_lc_charts(
    df: pd.DataFrame, col_name_t_y_m: str, col_name_n: str, col_name_q: str
) -> alt.HConcatChart:
    """An implementation that returns two charts to trend nominal and relative values

    chart 1-> bar trend chart
    chart 2-> stacked bar chart
    """
    chart1 = (
        alt.Chart(df)
        .mark_bar()
        .encode(x=f"{col_name_t_y_m}:O", y=f"{col_name_q}:Q", color=col_name_n)
    )

    chart2 = (
        alt.Chart(df)
        .mark_bar()
        .encode(
            x=f"{col_name_t_y_m}:O",
            y=alt.Y(f"{col_name_q}:Q", stack="normalize"),
            color=col_name_n,
        )
    )

    return alt.hconcat(chart1, chart2)



"""Taking inspiration from PyJanitor on testing Pandas Stuff"""

# Third party imports
import pytest

# chart_me imports
import chart_me.data_validation_strategy as vsc
import chart_me.errors as e


def test_column_does_not_exist_error(conftest_basic_dataframe):
    df = conftest_basic_dataframe
    c_validator = vsc.ValidateColumnStrategyDefault(df, "col_no_exist")
    with pytest.raises(e.ColumnDoesNotExistsError):
        c_validator.validate_column()


def test_column_is_all_null_error(conftest_basic_dataframe):
    df = conftest_basic_dataframe
    c_validator = vsc.ValidateColumnStrategyDefault(df, "all_nulls")
    with pytest.raises(e.ColumnAllNullError):
        c_validator.validate_column()


def test_column_is_mostly_null_error(conftest_basic_dataframe):
    df = conftest_basic_dataframe
    c_validator = vsc.ValidateColumnStrategyDefault(df, "mostly_nulls")
    with pytest.raises(e.ColumnTooManyNullsError):
        c_validator.validate_column()


@pytest.mark.parametrize(
    "col_names", ["inty_integers", "floaty_floats", "stringy_strings", "datie_dates"]
)
def test_columns_are_valid(conftest_basic_dataframe, col_names):
    df = conftest_basic_dataframe
    c_validator = vsc.ValidateColumnStrategyDefault(df, col_names)
    assert c_validator.validate_column()



"""Default implementation for single variable charting

This module contains all the logic to go from Metadata to actual charts
See supporting documentation that discusses the rules engine.

    Typical usage example:

    charts = assemble_univariate_charts(df, [col1], infered_data_types)
"""
# Standard library imports
from typing import List, Union

# Third party imports
import altair as alt
import pandas as pd

# chart_me imports
from chart_me.datatype_infer_strategy import ChartMeDataTypeMetaType, InferedDataTypes


def assemble_univariate_charts(
    df: pd.DataFrame, cols: List[str], infered_data_types: InferedDataTypes, **kwargs
) -> List[Union[alt.Chart, alt.HConcatChart]]:
    """Delegated Function to Manage Univariate Use Cases

    Args:
        df: dataframe
        cols: a single column name in a list
        infered_data_types: An instance of InferedDataTypes object

    Returns:
        List of altair charts or compounds charts

    Raises:
        ValueError if called with more then one column
    """
    if len(cols) != 1:
        raise ValueError("Only suport single column")
    col_name = cols[0]
    col_MT = infered_data_types.chart_me_data_types_meta[col_name]
    return_charts = []
    if col_MT == ChartMeDataTypeMetaType.QUANTITATIVE:
        return_charts.append(build_histogram(df, col_name))
    elif col_MT == ChartMeDataTypeMetaType.KEY:
        return_charts.append(build_hbar_agg(df.head(), col_name, "count()"))
    elif col_MT != ChartMeDataTypeMetaType.TEMPORAL:
        # chart_me imports
        from chart_me.pandas_util import pd_group_me

        agg_dict = {f"{col_name}": ["count"]}
        df = pd_group_me(df, col_name, agg_dict)
        return_charts.append(build_hbar_agg(df.head(), col_name, df.columns[-1]))
    elif col_MT == ChartMeDataTypeMetaType.TEMPORAL:
        # chart_me imports
        from chart_me.pandas_util import pd_group_me, pd_truncate_date

        temp_col_name = f"{col_name}_ym_temp"
        df[temp_col_name] = pd_truncate_date(df, col_name)
        agg_dict = {temp_col_name: ["count"]}
        df_a = pd_group_me(df, temp_col_name, agg_dict, is_temporal=True)
        print(df_a)
        print(df_a.columns)
        return_charts.append(build_vbar_agg(df_a, temp_col_name, df_a.columns[-1]))
        del df[temp_col_name]

    else:
        raise NotImplementedError(f"unknown handling of metatype-{str(col_MT)}")
    return return_charts


def build_hbar_agg(df: pd.DataFrame, col_name: str, agg_name: str) -> alt.Chart:
    """An implementation of horizontal bar chart"""
    chart = alt.Chart(df).mark_bar().encode(x=f"{agg_name}:Q", y=f"{col_name}:O")
    return chart


def build_vbar_agg(df: pd.DataFrame, col_name: str, agg_name: str) -> alt.Chart:
    """An implementation of vertical bar chart"""
    chart = alt.Chart(df).mark_bar().encode(x=f"{col_name}:O", y=f"{agg_name}:Q")
    return chart


def build_histogram(df: pd.DataFrame, col_name: str) -> alt.Chart:
    """An implementation of histogram chart"""
    chart = (
        alt.Chart(df).mark_bar().encode(alt.X(f"{col_name}:Q", bin=True), y="count()")
    )
    return chart



"""Imports core chart_me function for end users.

    Typical usage example:
    import chart_me as ce
    ce.chart_me(df, col1, col2...)
"""

__version__ = "0.1.4"


# chart_me imports
from chart_me.chart_me_logic import chart_me  # noqa: F401



"""Module defines a core enums, protocal and default  datatype infer strategy

Defines 2 Enum -> ChartMeDataType & ChartMeDataTypeMetaType the normalize
metadata. Additionally, defines the 'InferedDataTypes' that get referenced
downstream to the assembler

    Typical usage example:

    infer = InferDataTypeStrategyDefault.datatype_infer_strategy(df, cols)
    infer_dtypes = infer.infer_datatypes() #return an InferedDataTypes
"""

# Standard library imports
import sys
from dataclasses import dataclass
from typing import Dict, List, Optional

if sys.version_info >= (3, 8):
    # Standard library imports
    from typing import Protocol
else:
    # Third party imports
    from typing_extensions import Protocol

# TODO think about a reorganizing into seperate module
# Standard library imports
from enum import Enum

# Third party imports
import pandas as pd


class ChartMeDataType(Enum):
    """Defines Normalized Datatypes"""

    FLOATS = "F"  # -> Metrics Avg/Media, etc...
    INTEGER = "I"  # -> Hybrid Datatypes Numerical or seen As Categorical
    TEMPORAL = ("T",)  # -> this lead to 'O' otherwise use
    NOMINAL = ("N",)
    NOT_SUPPORTED_TYPE = "NA"


class ChartMeDataTypeMetaType(Enum):
    """Defines Normalized Metadata about Datatypes

    Defines Key, Boolean, Quantitative, Categorical High Cardinality,
    Categorical Low Cardinality, Temporal, Not Supported"""

    KEY = ("K",)  # -> map to CountD
    BOOLEAN = ("B",)  # -> Context Aware
    QUANTITATIVE = ("Q",)  # -> Floats
    CATEGORICAL_HIGH_CARDINALITY = ("C-HC",)  # top-k
    CATEGORICAL_LOW_CARDINALITY = ("C-LC",)
    TEMPORAL = ("T",)
    NOT_SUPPORTED_TYPE = "NA"


@dataclass
class InferedDataTypes:
    """Core Object Returned for Assembler - column level specifications"""

    preaggregated: bool
    chart_me_data_types: Dict[str, ChartMeDataType]
    chart_me_data_types_meta: Dict[str, ChartMeDataTypeMetaType]


class InferDataTypeStrategy(Protocol):
    """Defines protocol for InferDataTypeStrategy - requires infer_datatypes"""

    def infer_datatypes(self) -> InferedDataTypes:
        """Only Method Required"""
        raise NotImplementedError


class InferDataTypeStrategyDefault:
    """Default logic to specificy metadata to drive chart assembly logic"""

    def __init__(self, df: pd.DataFrame, cols: List[str], **kwargs):
        """set instances variable"""
        self.df = df
        self.cols = cols
        self.preaggregated: Optional[bool] = None
        self.__dict__.update(kwargs)
        self.col_to_cm_dtypes: Dict[str, ChartMeDataType] = {}
        self.col_to_cm_dtypes_meta: Dict[str, ChartMeDataTypeMetaType] = {}

    def _check_if_preaggregated_data(self, threshold: int = 100):
        """First check determines aggregation - influences data type"""
        if self.df.shape[0] <= threshold:
            # check if at least l column is completely unique
            self.preaggregated = (
                True if any(self.df.nunique() == self.df.shape[0]) else False
            )
        else:
            self.preaggregated = False
        return self.preaggregated

    def _get_raw_data_infer_type(self):
        """Apply *some* burden on user to setup DF accordingly"""
        # Third party imports
        from pandas.api.types import infer_dtype

        map_from_pd_cm = {
            "string": ChartMeDataType.NOMINAL,
            "bytes": ChartMeDataType.NOT_SUPPORTED_TYPE,
            "floating": ChartMeDataType.FLOATS,
            "integer": ChartMeDataType.INTEGER,
            "mixed-integer": ChartMeDataType.NOT_SUPPORTED_TYPE,  # TODO
            "mixed-integer-float": ChartMeDataType.FLOATS,
            "decimal": ChartMeDataType.FLOATS,
            "complex": ChartMeDataType.NOT_SUPPORTED_TYPE,
            "categorical": ChartMeDataType.NOMINAL,
            "boolean": ChartMeDataType.INTEGER,
            "datetime64": ChartMeDataType.TEMPORAL,
            "datetime": ChartMeDataType.TEMPORAL,
            "date": ChartMeDataType.TEMPORAL,
            "timedelta64": ChartMeDataType.NOMINAL,
            "timedelta": ChartMeDataType.NOMINAL,
            "time": ChartMeDataType.NOT_SUPPORTED_TYPE,
            "period": ChartMeDataType.NOMINAL,
            "mixed": ChartMeDataType.NOT_SUPPORTED_TYPE,  # inform user
            "unknown-array": ChartMeDataType.NOT_SUPPORTED_TYPE,
        }

        map_from_pd = self.df[self.cols].apply(infer_dtype).to_dict()
        self.col_to_cm_dtypes = {k: map_from_pd_cm[v] for k, v in map_from_pd.items()}
        return self.col_to_cm_dtypes

    # TODO add 'sampling' not processing over huge series
    def _calculate_override_data_infer_type(self):
        """Part to is evaluate"""
        for col, data_type in self.col_to_cm_dtypes.items():
            if data_type == ChartMeDataType.FLOATS:
                data_sans_nulls = self.df[col].dropna()
                if all(data_sans_nulls.apply(lambda x: x.is_integer())):
                    self.col_to_cm_dtypes[
                        col
                    ] = (
                        ChartMeDataType.INTEGER
                    )  # * Not casting becasause of pd.nan issue
            if data_type == ChartMeDataType.NOMINAL:
                # Third party imports
                from dateutil.parser import ParserError

                try:
                    self.df[col] = pd.to_datetime(self.df[col])
                    self.col_to_cm_dtypes[col] = ChartMeDataType.TEMPORAL
                except (ParserError, TypeError, ValueError):
                    pass
        return self.col_to_cm_dtypes

    def _get_data_infer_meta_type_col(self, col: str):
        """series of evaluations"""
        if self.col_to_cm_dtypes[col] == ChartMeDataType.NOT_SUPPORTED_TYPE:
            self.col_to_cm_dtypes_meta[col] = ChartMeDataTypeMetaType.NOT_SUPPORTED_TYPE
        if self.col_to_cm_dtypes[col] == ChartMeDataType.FLOATS:
            self.col_to_cm_dtypes_meta[col] = ChartMeDataTypeMetaType.QUANTITATIVE
        if self.col_to_cm_dtypes[col] == ChartMeDataType.TEMPORAL:
            self.col_to_cm_dtypes_meta[col] = ChartMeDataTypeMetaType.TEMPORAL
        if self.col_to_cm_dtypes[col] == ChartMeDataType.INTEGER:
            # Check if a Key - All Unique and No Nulls
            if self.df[col].nunique() == self.df[col].shape[0]:
                self.col_to_cm_dtypes_meta[col] = ChartMeDataTypeMetaType.KEY
            # Check if a Boolean Value
            elif (
                int(self.df[col].min()) == 0
                and int(self.df[col].max()) == 1
                and self.df[col].nunique() == 2
            ):
                self.col_to_cm_dtypes_meta[col] = ChartMeDataTypeMetaType.BOOLEAN
            else:
                self.col_to_cm_dtypes_meta[col] = ChartMeDataTypeMetaType.QUANTITATIVE
        if self.col_to_cm_dtypes[col] == ChartMeDataType.NOMINAL:

            hc_threshold = (
                self.__dict__.get("cardinality_threshold_ratio", 0.30)
                if self.preaggregated
                else self.__dict__.get("cardinality_threshold_ratio", 0.10)
            )
            # if 100; 10 or less low cardinality
            if self.df[col].nunique() == self.df[col].shape[0]:
                self.col_to_cm_dtypes_meta[col] = ChartMeDataTypeMetaType.KEY
            elif self.df[col].nunique() / self.df.shape[0] <= hc_threshold:
                self.col_to_cm_dtypes_meta[
                    col
                ] = ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY
            else:
                self.col_to_cm_dtypes_meta[
                    col
                ] = ChartMeDataTypeMetaType.CATEGORICAL_HIGH_CARDINALITY

    def _calculate_data_type_meta(self):
        """wrapper to iterate over all columns"""
        for col, val in self.col_to_cm_dtypes.items():
            self._get_data_infer_meta_type_col(col)
        return self.col_to_cm_dtypes_meta

    def infer_datatypes(self) -> InferedDataTypes:
        """Core function of stategy is returning metadata to rendering engine

        Args:
            df (pd.DataFrame): pandas dataframe
            cols (List[str]): list of columns passed in from user

        Returns:
            InferedDataTypes: describes MetaData to engine
        """
        pre_agg_flag = self._check_if_preaggregated_data()
        _ = self._get_raw_data_infer_type()
        infered_data_type = self._calculate_override_data_infer_type()
        infered_data_type_meta = self._calculate_data_type_meta()
        self.infered_data_types = InferedDataTypes(
            pre_agg_flag, infered_data_type, infered_data_type_meta
        )
        return self.infered_data_types



"""Collection of panda manipulations - leverage prior to charts

    This big idea is to keep pandas operations isolated from visuals
    keep Altair logic very simple if possible

"""
# Standard library imports
from typing import Dict, List, Union

# Third party imports
import pandas as pd


# TODO document function signature
def pd_group_me(
    df: pd.DataFrame,
    cols: Union[List[str], str],
    agg_dict: Dict,
    is_temporal: bool = False,
    make_long_form=False,
) -> pd.DataFrame:
    """A generic function to do group by aggregation in pandas

    helpful url: https://jamesrledoux.com/code/group-by-aggregate-pandas
    WARNING: Hard code logic to return var_name to "measures"

    Args:
        df: data
        cols: grouping columns
        agg_dict: aggregation dictionary: e.g. {'Age': ['mean', 'min', 'max']}
        is_temporal: boolean flag used to set 'order' by Dates versus Counts
        make_long_form: leverages reset_index and defaults

    Returns:
        pd.DataFrame: Returns tidy dataframe with default names
    """
    df = df.groupby(cols).agg(agg_dict).reset_index()
    key_cols = [f"{k}-{i}" for k in agg_dict.keys() for i in agg_dict[k]]
    sort_key = [key_cols][0]
    df.columns = (list([cols]) if isinstance(cols, str) else list(cols)) + list(
        key_cols
    )
    if is_temporal:
        df = df.sort_values(list([cols]), ascending=True)
    else:
        df = df.sort_values(sort_key, ascending=False)
    if make_long_form:
        df = pd.melt(
            df,
            id_vars=cols,
            var_name="measures",
            value_name=next(iter(agg_dict.keys())),
        )
        df["measures"] = df["measures"].str.split("-").str[-1]
    return df


def pd_truncate_date(df: pd.DataFrame, col: str) -> pd.Series:
    """Utility to make dates YY--MM--01 to Strings

    Helpful urls: https://predictivehacks.com/?all-tips=how-to-truncate-dates-to-month-in-pandas # noqa: E501
    Helpful urls: https://pandas.pydata.org/docs/reference/api/pandas.Series.dt.to_period.html # noqa: E501

    Args:
        df: dataframe
        col: column name of date to truncate

    Returns:
        pd.Series: returns a Series of "string" datatypes
    """
    return df[col].dt.to_period("M").astype(str)



# %%
# Third party imports
from interrogate import coverage

cov = coverage.InterrogateCoverage(paths=["src"])
results = cov.get_coverage()
cov.print_results(results, None, 1)
# %%
cov.print_results(results, None, 2)
# %%



# Third party imports
import pytest

# chart_me imports
import chart_me.datatype_infer_strategy as dti
from chart_me.datatype_infer_strategy import (
    ChartMeDataType,
    ChartMeDataTypeMetaType,
    InferedDataTypes,
)

# preagg_inputs = "cols,output"
# preagg_checks =


@pytest.mark.parametrize(
    "pa_cols, pa_output",
    [
        ("unique_vals", True),
        ("unique_with_nulls", False),
        ("duplicates", False),
        ("unique_vals,duplicates", True),
    ],
)
def test_preagg_checks(conftest_check_agg_dataframe, pa_cols, pa_output):
    cols = list(pa_cols.split(","))
    df = conftest_check_agg_dataframe[cols]
    infer = dti.InferDataTypeStrategyDefault(df, cols)
    assert infer._check_if_preaggregated_data() is pa_output


def test_preagg_checks_too_big(conftest_check_agg_dataframe_too_big):
    df = conftest_check_agg_dataframe_too_big
    cols = ["unique_vals"]
    infer = dti.InferDataTypeStrategyDefault(df, cols)
    assert infer._check_if_preaggregated_data() is False


def test_infer_data_types(ct_df_check_infer_dtypes):
    df = ct_df_check_infer_dtypes
    cols = ct_df_check_infer_dtypes.columns.tolist()
    # chart_me imports
    from chart_me.datatype_infer_strategy import InferDataTypeStrategyDefault

    infer = InferDataTypeStrategyDefault(df, cols)
    results = infer._get_raw_data_infer_type()
    print(results)
    expected = {
        "inty_integers": ChartMeDataType.INTEGER,
        "inty_integers_w_nulls": ChartMeDataType.FLOATS,
        "floaty_floats": ChartMeDataType.FLOATS,
        "stringy_strings": ChartMeDataType.NOMINAL,
        "stringy_strings_w_nulls": ChartMeDataType.NOMINAL,
        "datie_dates": ChartMeDataType.TEMPORAL,
        "datie_dates_null": ChartMeDataType.TEMPORAL,
        "boolie_bools": ChartMeDataType.INTEGER,
        "boolie_bools_null": ChartMeDataType.INTEGER,  # * This is interesting result
    }
    assert results == expected


def test_calculate_override_data_infer_type(ct_df_override_dtypes):
    df = ct_df_override_dtypes
    cols = df.columns.tolist()
    # chart_me imports
    from chart_me.datatype_infer_strategy import InferDataTypeStrategyDefault

    infer = InferDataTypeStrategyDefault(df, cols)
    _ = infer._get_raw_data_infer_type()
    results = infer._calculate_override_data_infer_type()
    expected = {
        "stringy_dates": ChartMeDataType.TEMPORAL,
        "inty_integers_w_nulls": ChartMeDataType.INTEGER,
        "floaty_integers": ChartMeDataType.INTEGER,
        "string_not_dates": ChartMeDataType.NOMINAL,
    }
    assert results == expected


def test_calculate_data_type_meta(ct_df_calculate_meta_dtypes):
    df = ct_df_calculate_meta_dtypes
    cols = df.columns.tolist()
    # chart_me imports
    from chart_me.datatype_infer_strategy import InferDataTypeStrategyDefault

    infer = InferDataTypeStrategyDefault(df, cols, cardinality_threshold_ratio=0.4)
    _ = infer._get_raw_data_infer_type()
    _ = infer._calculate_override_data_infer_type()
    results = infer._calculate_data_type_meta()
    expected = {
        "temporal_t": ChartMeDataTypeMetaType.TEMPORAL,
        "floaties_q": ChartMeDataTypeMetaType.QUANTITATIVE,
        "nominal_lc": ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY,
        "nominal_hc": ChartMeDataTypeMetaType.CATEGORICAL_HIGH_CARDINALITY,
        "nominal_k": ChartMeDataTypeMetaType.KEY,
        "integers_b": ChartMeDataTypeMetaType.BOOLEAN,
        "integers_k": ChartMeDataTypeMetaType.KEY,
        "integers_q": ChartMeDataTypeMetaType.QUANTITATIVE,
    }
    print(results)
    assert results == expected


def test_infer_datatype_function_has_protocol(ct_df_calculate_meta_dtypes):
    df = ct_df_calculate_meta_dtypes
    cols = df.columns.tolist()
    # chart_me imports
    from chart_me.datatype_infer_strategy import InferDataTypeStrategyDefault

    infer = InferDataTypeStrategyDefault(df, cols, cardinality_threshold_ratio=0.4)
    infer_type = infer.infer_datatypes()
    assert isinstance(infer_type, InferedDataTypes)


def test_infer_datatype_function_right_values(ct_df_calculate_meta_dtypes):
    df = ct_df_calculate_meta_dtypes
    cols = df.columns.tolist()
    # chart_me imports
    from chart_me.datatype_infer_strategy import InferDataTypeStrategyDefault

    infer = InferDataTypeStrategyDefault(df, cols, cardinality_threshold_ratio=0.4)
    infer_type = infer.infer_datatypes()
    expect = InferedDataTypes(
        preaggregated=True,
        chart_me_data_types={
            "temporal_t": ChartMeDataType.TEMPORAL,
            "integers_b": ChartMeDataType.INTEGER,
            "integers_q": ChartMeDataType.INTEGER,
            "integers_k": ChartMeDataType.INTEGER,
            "floaties_q": ChartMeDataType.FLOATS,
            "nominal_lc": ChartMeDataType.NOMINAL,
            "nominal_hc": ChartMeDataType.NOMINAL,
            "nominal_k": ChartMeDataType.NOMINAL,
        },
        chart_me_data_types_meta={
            "temporal_t": ChartMeDataTypeMetaType.TEMPORAL,
            "integers_b": ChartMeDataTypeMetaType.BOOLEAN,
            "integers_q": ChartMeDataTypeMetaType.QUANTITATIVE,
            "integers_k": ChartMeDataTypeMetaType.KEY,
            "floaties_q": ChartMeDataTypeMetaType.QUANTITATIVE,
            "nominal_lc": ChartMeDataTypeMetaType.CATEGORICAL_LOW_CARDINALITY,
            "nominal_hc": ChartMeDataTypeMetaType.CATEGORICAL_HIGH_CARDINALITY,
            "nominal_k": ChartMeDataTypeMetaType.KEY,
        },
    )
    print(infer_type)
    assert expect == infer_type



"""Defines default Chart Config"""

# Standard library imports
from dataclasses import dataclass
from typing import Type

# chart_me imports
from chart_me.charting_assembly_strategy import (
    AssembleChartsStrategy,
    AssembleChartsStrategyDefault,
)
from chart_me.data_validation_strategy import (
    ValidateColumnStrategy,
    ValidateColumnStrategyDefault,
)
from chart_me.datatype_infer_strategy import (
    InferDataTypeStrategy,
    InferDataTypeStrategyDefault,
)


@dataclass
class ChartConfig:
    """Default Instance of Chart Config"""

    validate_column_strategy: Type[
        ValidateColumnStrategy
    ] = ValidateColumnStrategyDefault
    datatype_infer_strategy: Type[InferDataTypeStrategy] = InferDataTypeStrategyDefault
    assemble_charts_strategy: Type[
        AssembleChartsStrategy
    ] = AssembleChartsStrategyDefault



"""Specify AssembleChartsStrategy Protocol; Default Implementation of Chart Assembly

The AssembleChartsStrategy Protocal requires one method - assemble_charts that return
a list of High Level Altair Charts or Compound Chart. The AssembleChartStrategy expects
two MetaData Objects InferedDataTypes that specificy metadata about columns in the
dataframe.

    Typical usage example:

    assembler = AssembleChartsStrategyDefault(df, cols, infer_dtypes)
    charts = assembler.assemble_charts()
"""
# Standard library imports
import sys
from typing import List, Union

if sys.version_info >= (3, 8):
    # Standard library imports
    from typing import Protocol
else:
    from typing_extensions import Protocol

# Standard library imports
import warnings

# Third party imports
import altair as alt
import pandas as pd

# chart_me imports
from chart_me.datatype_infer_strategy import ChartMeDataType, InferedDataTypes
from chart_me.errors import InsufficientValidColumnsError

from .bivariate import assemble_bivariate_charts
from .univariate import assemble_univariate_charts


class AssembleChartsStrategy(Protocol):
    """Defines the protocol type for AssembleChartsStrategy

    Protocal requires 1 method --- assemble_charts
    """

    def assemble_charts(self) -> List[Union[alt.Chart, alt.HConcatChart]]:
        """Defines Protocal Type for AssembleChartsStrategy

        Args:
            df (pd.DataFrame):
            cols (List[str]): at least 1 and no more then 2 at this time
            infered_data_types (InferedDataTypes): metadata required to guide viz logic

        Returns:
            List[Union[alt.Chart, alt.HConcatChart]]: return a list of Altair visuals
        """


class AssembleChartsStrategyDefault:
    """Default implementation of AssembleChartsStrategy

    Default implementation strategy provided by chart_me. The core logic is
    provided by seperate modules depending on number of columns provided - so far
    2 are build univariate & bivariate. More then 2 columns will throw an error
    """

    def __init__(
        self,
        df: pd.DataFrame,
        cols: List[str],
        infered_data_types: InferedDataTypes,
        **kwargs,
    ) -> None:
        """Init Default Assemble Charts Strategy Object

        Args:
            df: dataframe
            cols: column names at least 1 and no more then 2 at this time
            infered_data_types: An InferedDataTypes instance containing metadata
            required to guide Altair rules
        """
        self.df = df
        self.user_provided_cols = cols
        self.preaggreted_fl = infered_data_types.preaggregated
        self.infered_data_types = infered_data_types
        self.__dict__.update(kwargs)

    def assemble_charts(self) -> List[Union[alt.Chart, alt.HConcatChart]]:
        """assembles charts based on columns count

        Returns:
            A list of Altair Chart/Compound charts to display

        Raises:
            InsufficientValidColumnsError: Will occur if no columns available after
                removing unsupported types.
            NotImplementedError: Only support 1 or 2 columns at this time

        """

        self.supported_cols: List[Union[alt.Chart, alt.HConcatChart]] = []
        # logic is predicated on number of columns & preaggregated status
        for c in self.user_provided_cols:
            if (
                self.infered_data_types.chart_me_data_types[c]
                == ChartMeDataType.NOT_SUPPORTED_TYPE
            ):
                warnings.warn(f"{c}-is not a supported datatype - ignoring")
            else:
                self.supported_cols.append(c)
        if not len(self.supported_cols):
            raise InsufficientValidColumnsError(
                "There's no columns with supported DataType"
            )
        elif len(self.supported_cols) == 1:
            charts = assemble_univariate_charts(
                self.df, self.supported_cols, self.infered_data_types
            )
        elif len(self.supported_cols) == 2:
            charts = assemble_bivariate_charts(
                self.df, self.supported_cols, self.infered_data_types
            )
        else:
            raise NotImplementedError("Only support two columns at this time")
        return charts
